import * as mc from '@minecraft/server';
import * as ui from '@minecraft/server-ui';
import * as db from './database.js'

/*
LOGS FORMAT
[ timestamp, type, source, block, args[] ]
timestamp - action timestamp
source - player or entity that performed the action
type - action type: 0 blockExplode, 1 blockBreak, 2 blockPlaced, 3 itemRemoved, 4 itemAdded
block - source block in block action, item in itemTransfered

args - additional arguments for actions like itemTransfered:
itemTransfered = count
blockPlaced = properties
*/

const default_perms = {
    realmguard: [],
    inspect: [],
    rollback: [],
    permissions: []
}
const permissions_db = new db.database('rg_permissions', default_perms);
var permissions = permissions_db.get();
if (Object.keys(default_perms).length !== Object.keys(permissions).length) {
    permissions_db.set(default_perms)
}

var inspect = {};
var rollback = {};
var rollbackSelection = {};
var page = {}
var inventories = {
    blocks: {},
    players: {}
}
var playerState = {
    interaction: {},
    rotation: {}
}
const containers = [
    'minecraft:chest',
    'minecraft:trapped_chest',
    'minecraft:barrel',
    'minecraft:furnace',
    'minecraft:hopper',
    'minecraft:blast_furnace',
    'minecraft:smoker',
    'minecraft:beacon',
    'minecraft:brewing_stand',
    'minecraft:dispenser',
    'minecraft:dropper',
    'minecraft:lectern',
    'minecraft:jukebox'
];

mc.system.events.beforeWatchdogTerminate.subscribe((evt) => {
    evt.cancel = true;
})

function getDateTime(time) {
    let date = new Date(time);
    let month = date.getMonth() + 1
    return {
        year: date.getFullYear().toString(),
        month: month.toString().length == 2 ? month.toString() : `0${month}`,
        date: date.getDate().toString().length == 2 ? date.getDate().toString() : `0${date.getDate()}`,
        hours: date.getHours().toString().length == 2 ? date.getHours().toString() : `0${date.getHours()}`,
        minutes: date.getMinutes().toString().length == 2 ? date.getMinutes().toString() : `0${date.getMinutes()}`,
        seconds: date.getSeconds().toString().length == 2 ? date.getSeconds().toString() : `0${date.getSeconds()}`
    }
}

function getProperties(block) {
    if (block.typeId.split(':')[0] !== 'minecraft') {
        return;
    }
    let output = {}
    let permutation = block.permutation;
    for (let property of permutation.getAllProperties()) {
        output[property.name] = property.value;
    }
    return output;
}

function setProperties(block, properties) {
    let permutation = block.permutation;
    for (let key in properties) {
        let property = permutation.getProperty(key);
        property.value = properties[key]
    }
    block.setPermutation(permutation);
}

async function log(init, record, block) {
    let log = new db.log(packLocation(block), [[1321620000000].concat(init)]);
    let array = log.get();
    array.push([Date.now()].concat(record));
    await log.set(array);
    return true;
}

function showLogs(block, player) {
    if (page[player.name] === undefined) {
        page[player.name] = 0;
    }
    let form = new ui.ActionFormData();
    let buttons = [];
    form.title(`Block ${block.x} ${block.y} ${block.z}, page ${page[player.name]}`);

    let log = new db.log(packLocation(block), [[1321620000000, 2, "world", block.typeId]]).get();
    let content = {
        start: (log.length - (page[player.name] + 1) * 10 - 1) < 0 ? 0 : (log.length - (page[player.name] + 1) * 10 - 1),
        end: log.length - (page[player.name]) * 10
    }
    let records = log.slice(content.start, content.end).reverse();
    let body = '';
    for (let record of records) {
        let date = getDateTime(record[0]);
        let recordInfo;
        let recordDate = `§5${date.year}/${date.month}/${date.date} ${date.hours}:${date.minutes}:${date.seconds}`;
        if (record[1] == 0) {
            recordInfo = `§3${record[2]} §rexplode §3${record[3]}`
        } else if (record[1] == 1) {
            recordInfo = `§3${record[2]} §rbroke §3${record[3]}`
        } else if (record[1] == 2) {
            recordInfo = `§3${record[2]} §rplaced §3${record[3]}`
        } else if (record[1] == 3) {
            recordInfo = `§3${record[2]} §rremoved §3${record[4]}x ${record[3]}`
        } else if (record[1] == 4) {
            recordInfo = `§3${record[2]} §radded §3${record[4]}x ${record[3]}`
        }
        body = body + '\n' + recordDate + ' ' + recordInfo
    }
    form.body(body);
    if (content.end !== log.length) {
        form.button('<<<Prev')
        buttons.push('prev');
    }
    if (content.start !== 0) {
        form.button('Next>>>')
        buttons.push('next');
    }
    form.button('Exit')
    buttons.push('exit');


    form.show(player).then((r) => {
        if (r.canceled) {
            delete page[player.name];
            return;
        }
        let button = buttons[r.selection];
        if (button == 'prev') {
            page[player.name]--;
            showLogs(block, player);
        } else if (button == 'next') {
            page[player.name]++;
            showLogs(block, player);
        }
    })
}

function createRollback(player) {
    //Setting up rollback
    let positions = rollback[player.name];
    if (positions[0].length < 3 || positions[1].length < 3) {
        player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cYou need to select a cuboid to perform a rollback."}]}`)
        return;
    }

    let date = getDateTime(Date.now());
    let form = new ui.ModalFormData();
    form.title('Rollback setup');
    form.textField('Year:', 'Year', date.year).textField('Month:', 'Month', date.month).textField('Date:', 'Date', date.date);
    form.textField('Hour:', 'Hour', date.hours).textField('Minute:', 'Minute', date.minutes).textField('Second:', 'Second', date.seconds);
    form.show(player).then((r) => {
        //testing form
        let input = r.formValues;
        if (r.canceled || r.formValues.length < 6) {
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cYou need to specify a full date-time to perform a rollback."}]}`)
            return;
        }
        for (let i = 0; i < 6; i++) {
            if (i == 0) {
                if (!(input[i].length == 4)) {
                    player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cInvalid date."}]}`)
                    return;
                }
            } else {
                if (!(input[i].length == 2)) {
                    player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cInvalid date."}]}`)
                    return;
                }
            }
        }

        //rollback
        let time = new Date(input[0], input[1] - 1, input[2], input[3], input[4], input[5]).getTime()
        rollbackLoop(positions, player, time)
    })
}

async function rollbackLoop(positions, player, time) {
    let console = new db.Console(player);
    let pos1 = [Math.min(positions[0][0], positions[1][0]), Math.min(positions[0][1], positions[1][1]), Math.min(positions[0][2], positions[1][2])]
    let pos2 = [Math.max(positions[0][0], positions[1][0]), Math.max(positions[0][1], positions[1][1]), Math.max(positions[0][2], positions[1][2])]
    let size = {
        x: pos2[0] - pos1[0],
        y: pos2[1] - pos1[1],
        z: pos2[2] - pos1[2]

    }
    let chunkSize = 8192;
    let totalBlocks = size.x * size.y * size.z;
    let totalChunks = Math.ceil(totalBlocks / chunkSize);
    let blocks = 0;
    let chunk = 0;
    for (let x = 0; x <= size.x; x++) {
        for (let y = 0; y <= size.y; y++) {
            for (let z = 0; z <= size.z; z++) {
                if (blocks > chunkSize) {
                    blocks = 0;
                    await console.run(`title @s actionbar §3${chunk}/${totalChunks}`);
                    chunk++;
                }
                let block = player.dimension.getBlock(new mc.BlockLocation(pos1[0] + x, pos1[1] + y, pos1[2] + z));
                rollbackBlock(block, time);
                blocks++;
            }
        }
    }
    player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§dRollback completed."}]}`)
}

async function rollbackBlock(block, time) {
    let logId = packLocation(block);
    if (mc.world.scoreboard.getObjective(logId) !== null) {
        let log = new db.log(packLocation(block), [[0]]).get();
        let rollbackState;
        if (log[0][0] > time) {
            rollbackState = log[0][0]
        } else {
            for (let record of log.reverse()) {
                if (record[0] < time) {
                    rollbackState = record;
                    break;
                }
            }
        }
        if (rollbackState[1] == 0 && rollbackState[1] == 1) {
            block.setType(mc.MinecraftBlockTypes.air);
        } else if (rollbackState[1] == 2) {
            block.setType(mc.MinecraftBlockTypes.get(rollbackState[3]));
            if (Object.keys(rollbackState[4]).length > 0) {
                setProperties(block, rollbackState[4])
            }
        }
    }
    return true;
}

function packLocation(block) {
    return [block.x, block.y, block.z].join('.')
}

function parseInventory(container) {
    let obj = {};
    for (let slot = 0; slot < container.size; slot++) {
        let item = container.getItem(slot);
        if (item) {
            let id = item.typeId;
            if (id in obj) {
                obj[id] += item.amount;
            } else {
                obj[id] = item.amount;
            }
        }
    }
    return obj;
}

function joinKeys(obj1, obj2) {
    let result = [];
    for (let key of Object.keys(obj1)) {
        if (!(key in obj2)) result.push(key);
    }
    return Object.keys(obj2).concat(result);
}

async function inventoryLog(player = { before: {}, after: {} }, block = { before: {}, after: {} }, name, sourceBlock) {
    for (let item of joinKeys(player.before, player.after)) {
        let playerAmount = { before: player.before[item] ?? 0, after: player.after[item] ?? 0 };
        let blockAmount = { before: block.before[item] ?? 0, after: block.after[item] ?? 0 };
        if ((playerAmount.before < playerAmount.after) && (blockAmount.before > blockAmount.after)) {
            //item removed
            let amount = blockAmount.before - blockAmount.after;
            await log([2, "world", 'minecraft:chest'], [3, name, item, amount], sourceBlock)
        } else if ((playerAmount.before > playerAmount.after) && (blockAmount.before < blockAmount.after)) {
            //item added
            let amount = blockAmount.after - blockAmount.before;
            await log([2, "world", 'minecraft:chest'], [4, name, item, amount], sourceBlock)
        }
    }
}

//Explosion Handler
mc.world.events.beforeExplosion.subscribe((evt) => {
    evt.impactedBlocks.forEach((blockLocation) => {
        let block = evt.dimension.getBlock(blockLocation);
        log([2, "world", block.typeId, getProperties(block)], [0, evt.source.typeId, block.typeId], block)
    });
})

//BlockBreak Handler
mc.world.events.blockBreak.subscribe((evt) => {
    let blockId = evt.brokenBlockPermutation.type.id;
    log([2, "world", blockId, getProperties(evt.block)], [1, evt.player.name, blockId], evt.block)
})

//BlockPlace Handler
mc.world.events.blockPlace.subscribe((evt) => {
    log([2, "world", 'minecraft:air'], [2, evt.player.name, evt.block.typeId, getProperties(evt.block)], evt.block)
})

//Chest Test
mc.world.events.beforeItemUseOn.subscribe((evt) => {
    let player = evt.source;
    let block = evt.source.dimension.getBlock(evt.blockLocation);

    if (containers.indexOf(block.typeId) !== -1) {
        if (!(block.getComponent('inventory'))) return;
        playerState.interaction[player.nameTag] = block;
        playerState.rotation[player.nameTag] = [Math.floor(player.rotation.x / 25), Math.floor(player.rotation.y / 25)];
        inventories.players[player.nameTag] = parseInventory(player.getComponent('inventory').container);
        inventories.blocks[packLocation(block)] = parseInventory(block.getComponent('inventory').container);
    }
})

//ItemTransfer Handler
mc.system.runSchedule(() => {
    mc.world.getAllPlayers().forEach((player) => {
        if (!(player.name in playerState.interaction)) {
            return;
        }
        //waiting
        let rotation = playerState.rotation[player.name]
        let currentRotation = [Math.floor(player.rotation.x / 25), Math.floor(player.rotation.y / 25)];
        if (rotation[0] == currentRotation[0] && rotation[1] == currentRotation[1]) {
            return;
        }
        //handling
        let playerInventoryNew = parseInventory(player.getComponent('inventory').container);
        let playerInventoryOld = inventories.players[player.name];

        let block = playerState.interaction[player.name];
        let blockInventoryNew = parseInventory(block.getComponent('inventory').container);
        let blockInventoryOld = inventories.blocks[packLocation(block)];

        inventoryLog({
            before: playerInventoryOld, after: playerInventoryNew
        }, {
            before: blockInventoryOld, after: blockInventoryNew
        },
            player.name, block)

        delete playerState.interaction[player.name];
        delete playerState.rotation[player.name];
        delete inventories.blocks[packLocation(block)];
        delete inventories.players[player.name];
    })
}, 5)

//Inspection Handler
mc.world.events.beforeItemUseOn.subscribe((evt) => {
    let player = evt.source;
    let block = evt.source.dimension.getBlock(evt.blockLocation);
    if (!(player.name in inspect)) return;
    evt.cancel = true;

    if (packLocation(block) === inspect[player.name]) return;
    inspect[player.name] = packLocation(block);

    showLogs(block, player)
})

//Rollback Position Handler
mc.world.events.beforeItemUseOn.subscribe((evt) => {
    let player = evt.source;
    let block = evt.source.dimension.getBlock(evt.blockLocation);
    if (!(player.name in rollback)) return;
    evt.cancel = true;

    if (rollbackSelection[player.name] > Date.now()) return;
    rollbackSelection[player.name] = Date.now() + 500;

    if (player.isSneaking) {
        rollback[player.name][1] = [block.x, block.y, block.z];
        player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§dSecond position set"}]}`)
    } else {
        rollback[player.name][0] = [block.x, block.y, block.z];
        player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§dFirst position set"}]}`)
    }
})

//Rollback Handler
mc.world.events.itemUse.subscribe((evt) => {
    let player = evt.source;
    if (!(player.name in rollback)) return;
    if (rollbackSelection[player.name] > Date.now()) return;

    createRollback(player);
})

//COMMANDS
const commands = {
    inspect: (player, args) => {
        if (!(rollback[player.name] === undefined)) {
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cYou can't enable inspector mode while in rollback mode."}]}`)
            return;
        }
        if (inspect[player.name] === undefined) {
            inspect[player.name] = 0
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§dInspector mode now enabled."}]}`)
        } else {
            delete inspect[player.name];
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§dInspector mode now disabled."}]}`)
        }
    },
    rollback: (player, args) => {
        if (!(inspect[player.name] === undefined)) {
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cYou can't enable rollback mode while in inspector mode."}]}`)
            return;
        }
        if (rollback[player.name] === undefined) {
            rollback[player.name] = [[], []]
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§dRollback mode now enabled."}]}`)
        } else {
            delete rollback[player.name];
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§dRollback mode now disabled."}]}`)
        }
    },
    permissions: (player, args) => {
        if (!(args[4])) {
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§6Usage: ${args[0]} ${args[1]} <target/everyone> <add/remove> <subcommand/realmguard>"}]}`)
            return;
        }
        if (!(args[4] in permissions)) {
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cSubcommand ${args[4]} does not exists. Consider using full command instead of alias."}]}`)
            return;
        }
        let idx = permissions[args[4]].indexOf(args[2]);
        if (args[3] == 'add') {
            if (idx !== -1) {
                player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§c${args[2]} already has permission to use ${args[4]}"}]}`)
                return;
            }
            permissions[args[4]].push(args[2]);
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§dAdded permission to use ${args[4]} for ${args[2]}"}]}`)
        } else if (args[3] == 'remove') {
            if (idx === -1) {
                player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§c${args[2]} don't have permission to use ${args[4]}"}]}`)
                return;
            }
            permissions[args[4]].splice(idx, 1);
            player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§dRemoved permission to use ${args[4]} for ${args[2]}"}]}`)
        }
        permissions_db.set(permissions);
    }
}

const aliases = {
    i: commands.inspect,
    perms: commands.permissions,
    p: commands.permissions,
    r: commands.rollback
}

mc.world.events.beforeChat.subscribe((evt) => {
    let args = evt.message.split(' ');
    const player = evt.sender
    if (args[0] !== '.rg' && args[0] !== '.realmguard') return;
    evt.cancel = true;
    if (!(args[1])) {
        player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"---§dRealmGuard Command List§r---"}]}`)
        player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"inspect - Enable/Disable Inspector mode"}]}`)
        player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"rollback - Enable/Disable Rollback mode"}]}`)
        player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"permissions - Add/Remove permissions to use RealmGuard commands for specific players"}]}`)
        return;
    }
    if (!(args[1] in commands || args[1] in aliases)) {
        player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cUnknown subcommand: ${args[1]}."}]}`)
        return;
    };
    let command = args[1] in commands ? args[1] : aliases[args[1]].name;
    if (!(permissions[command].indexOf(player.name) !== -1 || permissions['realmguard'].indexOf(player.name) !== -1)
        && !(permissions[command].indexOf('everyone') !== -1 || permissions['realmguard'].indexOf('everyone') !== -1)
        && !(player.isOp())) {
        player.runCommandAsync(`tellraw @s {"rawtext":[{"text":"§cYou don't have permission to use this command"}]}`)
        return;
    }
    commands[command](player, args);
})